@@ -85,7 +85,7 @@ encode_mouse_event_impl(unsigned int x, unsigned int y, int mouse_tracking_proto
 
 static int
 encode_mouse_event(Window *w, int button, MouseAction action, int mods) {
-    unsigned int x = w->mouse_cell_x + 1, y = w->mouse_cell_y + 1; // 1 based indexing
+    unsigned int x = w->mouse_pos.cell_x + 1, y = w->mouse_pos.cell_y + 1; // 1 based indexing
     Screen *screen = w->render_data.screen;
     return encode_mouse_event_impl(x, y, screen->modes.mouse_tracking_protocol, button, action, mods);
 
@@ -142,6 +142,7 @@ cell_for_pos(Window *w, unsigned int *x, unsigned int *y, OSWindow *os_window) {
         mouse_x = MIN(MAX(mouse_x, left), right);
         mouse_y = MIN(MAX(mouse_y, top), bottom);
     }
+    w->mouse_pos.x = mouse_x - left; w->mouse_pos.y = mouse_y - top;
     if (mouse_x < left || mouse_y < top || mouse_x > right || mouse_y > bottom) return false;
     if (mouse_x >= g->right) qx = screen->columns - 1;
     else if (mouse_x >= g->left) qx = (unsigned int)((double)(mouse_x - g->left) / os_window->fonts_data->cell_width);
@@ -164,14 +165,14 @@ update_drag(bool from_button, Window *w, bool is_release, int modifiers) {
             global_state.active_drag_in_window = 0;
             w->last_drag_scroll_at = 0;
             if (screen->selection.in_progress)
-                screen_update_selection(screen, w->mouse_cell_x, w->mouse_cell_y, true);
+                screen_update_selection(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y, true);
         }
         else {
             global_state.active_drag_in_window = w->id;
-            screen_start_selection(screen, w->mouse_cell_x, w->mouse_cell_y, modifiers == (int)OPT(rectangle_select_modifiers) || modifiers == ((int)OPT(rectangle_select_modifiers) | GLFW_MOD_SHIFT), EXTEND_CELL);
+            screen_start_selection(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y, modifiers == (int)OPT(rectangle_select_modifiers) || modifiers == ((int)OPT(rectangle_select_modifiers) | GLFW_MOD_SHIFT), EXTEND_CELL);
         }
     } else if (screen->selection.in_progress) {
-        screen_update_selection(screen, w->mouse_cell_x, w->mouse_cell_y, false);
+        screen_update_selection(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y, false);
     }
 }
 
@@ -179,10 +180,7 @@ update_drag(bool from_button, Window *w, bool is_release, int modifiers) {
 bool
 drag_scroll(Window *w, OSWindow *frame) {
     unsigned int margin = frame->fonts_data->cell_height / 2;
-    double left = window_left(w, frame), top = window_top(w, frame), right = window_right(w, frame), bottom = window_bottom(w, frame);
-    double x = frame->mouse_x, y = frame->mouse_y;
-    if (y < top || y > bottom) return false;
-    if (x < left || x > right) return false;
+    double y = frame->mouse_y;
     bool upwards = y <= (w->geometry.top + margin);
     if (upwards || y >= w->geometry.bottom - margin) {
         Screen *screen = w->render_data.screen;
@@ -204,37 +202,62 @@ static inline void
 extend_selection(Window *w) {
     Screen *screen = w->render_data.screen;
     if (screen_has_selection(screen)) {
-        screen_update_selection(screen, w->mouse_cell_x, w->mouse_cell_y, false);
+        screen_update_selection(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y, false);
     }
 }
 
 static inline void
-extend_url(Screen *screen, Line *line, index_type *x, index_type *y) {
+extend_url(Screen *screen, Line *line, index_type *x, index_type *y, char_type sentinel) {
     unsigned int count = 0;
     while(count++ < 10) {
         if (*x != line->xnum - 1) break;
         line = screen_visual_line(screen, *y + 1);
         if (!line) break; // we deliberately allow non-continued lines as some programs, like mutt split URLs with newlines at line boundaries
-        index_type new_x = line_url_end_at(line, 0, false);
+        index_type new_x = line_url_end_at(line, 0, false, sentinel);
         if (!new_x) break;
         *y += 1; *x = new_x;
     }
 }
 
+static inline char_type
+get_url_sentinel(Line *line, index_type url_start) {
+    char_type before = 0, sentinel;
+    if (url_start > 0 && url_start < line->xnum) before = line->cpu_cells[url_start - 1].ch;
+    switch(before) {
+        case '"':
+        case '\'':
+        case '*':
+            sentinel = before; break;
+        case '(':
+            sentinel = ')'; break;
+        case '[':
+            sentinel = ']'; break;
+        case '{':
+            sentinel = '}'; break;
+        case '<':
+            sentinel = '>'; break;
+        default:
+            sentinel = 0; break;
+    }
+    return sentinel;
+}
+
 static inline void
 detect_url(Screen *screen, unsigned int x, unsigned int y) {
     bool has_url = false;
     index_type url_start, url_end = 0;
     Line *line = screen_visual_line(screen, y);
+    char_type sentinel;
     if (line) {
         url_start = line_url_start_at(line, x);
-        if (url_start < line->xnum) url_end = line_url_end_at(line, x, true);
+        sentinel = get_url_sentinel(line, url_start);
+        if (url_start < line->xnum) url_end = line_url_end_at(line, x, true, sentinel);
         has_url = url_end > url_start;
     }
     if (has_url) {
         mouse_cursor_shape = HAND;
         index_type y_extended = y;
-        extend_url(screen, line, &url_end, &y_extended);
+        extend_url(screen, line, &url_end, &y_extended, sentinel);
         screen_mark_url(screen, url_start, y, url_end, y_extended);
     } else {
         mouse_cursor_shape = BEAM;
@@ -254,8 +277,8 @@ HANDLER(handle_move_event) {
     if (!cell_for_pos(w, &x, &y, global_state.callback_os_window)) return;
     Screen *screen = w->render_data.screen;
     detect_url(screen, x, y);
-    bool mouse_cell_changed = x != w->mouse_cell_x || y != w->mouse_cell_y;
-    w->mouse_cell_x = x; w->mouse_cell_y = y;
+    bool mouse_cell_changed = x != w->mouse_pos.cell_x || y != w->mouse_pos.cell_y;
+    w->mouse_pos.cell_x = x; w->mouse_pos.cell_y = y;
     bool handle_in_kitty = (
             (screen->modes.mouse_tracking_mode == ANY_MODE ||
             (screen->modes.mouse_tracking_mode == MOTION_MODE && button >= 0)) &&
@@ -282,14 +305,14 @@ multi_click(Window *w, unsigned int count) {
     index_type start, end;
     bool found_selection = false;
     SelectionExtendMode mode = EXTEND_CELL;
-    unsigned int y1 = w->mouse_cell_y, y2 = w->mouse_cell_y;
+    unsigned int y1 = w->mouse_pos.cell_y, y2 = w->mouse_pos.cell_y;
     switch(count) {
         case 2:
-            found_selection = screen_selection_range_for_word(screen, w->mouse_cell_x, &y1, &y2, &start, &end);
+            found_selection = screen_selection_range_for_word(screen, w->mouse_pos.cell_x, &y1, &y2, &start, &end);
             mode = EXTEND_WORD;
             break;
         case 3:
-            found_selection = screen_selection_range_for_line(screen, w->mouse_cell_y, &start, &end);
+            found_selection = screen_selection_range_for_line(screen, w->mouse_pos.cell_y, &start, &end);
             mode = EXTEND_LINE;
             break;
         default:
@@ -301,19 +324,39 @@ multi_click(Window *w, unsigned int count) {
     }
 }
 
+static inline double
+distance(double x1, double y1, double x2, double y2) {
+    return sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2));
+}
+
 HANDLER(add_click) {
     ClickQueue *q = &w->click_queue;
     if (q->length == CLICK_QUEUE_SZ) { memmove(q->clicks, q->clicks + 1, sizeof(Click) * (CLICK_QUEUE_SZ - 1)); q->length--; }
     double now = monotonic();
 #define N(n) (q->clicks[q->length - n])
-    N(0).at = now; N(0).button = button; N(0).modifiers = modifiers;
+    N(0).at = now; N(0).button = button; N(0).modifiers = modifiers; N(0).x = w->mouse_pos.x; N(0).y = w->mouse_pos.y;
     q->length++;
+    double multi_click_allowed_radius = 1.2 * global_state.callback_os_window->fonts_data->cell_height;
     // Now dispatch the multi-click if any
-    if (q->length > 2 && N(1).at - N(3).at <= 2 * OPT(click_interval)) {
-        multi_click(w, 3);
-        q->length = 0;
-    } else if (q->length > 1 && N(1).at - N(2).at <= OPT(click_interval)) {
-        multi_click(w, 2);
+    if (q->length > 2) {
+        // possible triple-click
+        if (
+                N(1).at - N(3).at <= 2 * OPT(click_interval) &&
+                distance(N(1).x, N(1).y, N(3).x, N(3).y) <= multi_click_allowed_radius
+           ) {
+            multi_click(w, 3);
+            q->length = 0;
+        }
+    }
+    if (q->length > 1) {
+        // possible double-click
+        if (
+                N(1).at - N(2).at <= OPT(click_interval) &&
+                distance(N(1).x, N(1).y, N(2).x, N(2).y) <= multi_click_allowed_radius
+           ) {
+            multi_click(w, 2);
+        }
+
     }
 #undef N
 }
@@ -321,6 +364,7 @@ HANDLER(add_click) {
 static inline void
 open_url(Window *w) {
     Screen *screen = w->render_data.screen;
+    detect_url(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y);
     screen_open_url(screen);
 }
 
@@ -452,23 +496,43 @@ focus_in_event() {
 }
 
 void
-mouse_event(int button, int modifiers) {
+enter_event() {
+#ifdef __APPLE__
+    // On cocoa there is no way to configure the window manager to
+    // focus windows on mouse enter, so we do it ourselves
+    if (OPT(focus_follows_mouse) && !global_state.callback_os_window->is_focused) {
+        focus_os_window(global_state.callback_os_window, false);
+    }
+#endif
+}
+
+void
+mouse_event(int button, int modifiers, int action) {
     MouseShape old_cursor = mouse_cursor_shape;
     bool in_tab_bar;
     unsigned int window_idx = 0;
     Window *w = NULL;
-    if (button == -1 && global_state.active_drag_in_window) {  // drag move
-        w = window_for_id(global_state.active_drag_in_window);
-        if (w) {
-            button = currently_pressed_button();
-            if (button == GLFW_MOUSE_BUTTON_LEFT) {
-                clamp_to_window = true;
-                handle_move_event(w, button, modifiers, window_idx);
-                clamp_to_window = false;
-                return;
+    if (global_state.active_drag_in_window) {
+        if (button == -1) {  // drag move
+            w = window_for_id(global_state.active_drag_in_window);
+            if (w) {
+                button = currently_pressed_button();
+                if (button == GLFW_MOUSE_BUTTON_LEFT) {
+                    clamp_to_window = true;
+                    Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
+                    for (window_idx = 0; window_idx < t->num_windows && t->windows[window_idx].id != w->id; window_idx++);
+                    handle_move_event(w, button, modifiers, window_idx);
+                    clamp_to_window = false;
+                    return;
+                }
+            }
+        }
+        else if (action == GLFW_RELEASE && button == GLFW_MOUSE_BUTTON_LEFT) {
+            w = window_for_id(global_state.active_drag_in_window);
+            if (w) {
+                update_drag(true, w, true, modifiers);
             }
         }
-
     }
     w = window_for_event(&window_idx, &in_tab_bar);
     if (in_tab_bar) {
@@ -491,27 +555,92 @@ mouse_event(int button, int modifiers) {
 }
 
 void
-scroll_event(double UNUSED xoffset, double yoffset) {
-    // glfw inverts the y-axis when reporting scroll events under wayland
-    // Until this is fixed in upstream, invert y ourselves.
-    if (global_state.is_wayland) yoffset *= -1;
-    int s = (int) round(yoffset * OPT(wheel_scroll_multiplier));
-    if (s == 0) return;
-    bool upwards = s > 0;
+scroll_event(double UNUSED xoffset, double yoffset, int flags) {
     bool in_tab_bar;
+    static id_type window_for_momentum_scroll = 0;
+    static bool main_screen_for_momentum_scroll = false;
     unsigned int window_idx = 0;
     Window *w = window_for_event(&window_idx, &in_tab_bar);
-    if (w) {
-        Screen *screen = w->render_data.screen;
-        if (screen->linebuf == screen->main_linebuf) {
-            screen_history_scroll(screen, abs(s), upwards);
-        } else {
-            if (screen->modes.mouse_tracking_mode) {
-                int sz = encode_mouse_event(w, upwards ? GLFW_MOUSE_BUTTON_4 : GLFW_MOUSE_BUTTON_5, PRESS, 0);
-                if (sz > 0) { mouse_event_buf[sz] = 0; write_escape_code_to_child(screen, CSI, mouse_event_buf); }
-            } else {
-                fake_scroll(abs(s), upwards);
+    if (!w && !in_tab_bar) {
+        // allow scroll events even if window is not currently focused (in
+        // which case on some platforms such as macOS the mouse location is zeroed so
+        // window_for_event() does not work).
+        Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
+        if (t) w = t->windows + t->active_window;
+    }
+    if (!w) return;
+    Screen *screen = w->render_data.screen;
+
+    enum MomentumData { NoMomentumData, MomentumPhaseBegan, MomentumPhaseStationary, MomentumPhaseActive, MomentumPhaseEnded, MomentumPhaseCancelled, MomentumPhaseMayBegin };
+    enum MomentumData momentum_data = (flags >> 1) & 7;
+
+    switch(momentum_data) {
+        case NoMomentumData:
+            break;
+        case MomentumPhaseBegan:
+            window_for_momentum_scroll = w->id;
+            main_screen_for_momentum_scroll = screen->linebuf == screen->main_linebuf;
+            break;
+        case MomentumPhaseStationary:
+        case MomentumPhaseActive:
+            if (window_for_momentum_scroll != w->id || main_screen_for_momentum_scroll != (screen->linebuf == screen->main_linebuf)) return;
+            break;
+        case MomentumPhaseEnded:
+        case MomentumPhaseCancelled:
+            window_for_momentum_scroll = 0;
+            break;
+        case MomentumPhaseMayBegin:
+        default:
+            break;
+    }
+    if (yoffset == 0.0) return;
+
+    int s;
+    bool is_high_resolution = flags & 1;
+
+    if (is_high_resolution) {
+        yoffset *= OPT(touch_scroll_multiplier);
+        if (yoffset * global_state.callback_os_window->pending_scroll_pixels < 0) {
+            global_state.callback_os_window->pending_scroll_pixels = 0;  // change of direction
+        }
+        double pixels = global_state.callback_os_window->pending_scroll_pixels + yoffset;
+        if (fabs(pixels) < global_state.callback_os_window->fonts_data->cell_height) {
+            global_state.callback_os_window->pending_scroll_pixels = pixels;
+            return;
+        }
+        s = (int)round(pixels) / (int)global_state.callback_os_window->fonts_data->cell_height;
+        global_state.callback_os_window->pending_scroll_pixels = pixels - s * (int) global_state.callback_os_window->fonts_data->cell_height;
+    } else {
+        if (screen->linebuf == screen->main_linebuf || !screen->modes.mouse_tracking_mode) {
+            // Only use wheel_scroll_multiplier if we are scrolling kitty scrollback or in mouse
+            // tracking mode, where the application is responsible for interpreting scroll events
+            yoffset *= OPT(wheel_scroll_multiplier);
+        } else if (OPT(wheel_scroll_multiplier) < 0) {
+            // ensure that changing scroll direction still works, even though
+            // we are not using wheel_scroll_multiplier
+            yoffset *= -1;
+        }
+        s = (int) round(yoffset);
+        // apparently on cocoa some mice generate really small yoffset values
+        // when scrolling slowly https://github.com/kovidgoyal/kitty/issues/1238
+        if (s == 0 && yoffset != 0) s = yoffset > 0 ? 1 : -1;
+        global_state.callback_os_window->pending_scroll_pixels = 0;
+    }
+    if (s == 0) return;
+    bool upwards = s > 0;
+    if (screen->linebuf == screen->main_linebuf) {
+        screen_history_scroll(screen, abs(s), upwards);
+    } else {
+        if (screen->modes.mouse_tracking_mode) {
+            int sz = encode_mouse_event(w, upwards ? GLFW_MOUSE_BUTTON_4 : GLFW_MOUSE_BUTTON_5, PRESS, 0);
+            if (sz > 0) {
+                mouse_event_buf[sz] = 0;
+                for (s = abs(s); s > 0; s--) {
+                    write_escape_code_to_child(screen, CSI, mouse_event_buf);
+                }
             }
+        } else {
+            fake_scroll(abs(s), upwards);
         }
     }
 }
